from django_logic import Process, Transition
from django_logic.tasks import SideEffectTasks


class CeleryTransition(Transition):
    side_effects = SideEffectTasks()


class InvoiceProcess(Process):
    states = (
        ('draft', 'Draft'),
        ('paid', 'Paid'),
        ('void', 'Void'),
    )

    transitions = [
        Transition(action_name='approve',sources=['draft'], target='approved'),
        CeleryTransition(action_name='send_to_customer',
                         sources=['approved'],
                         side_effects=['app.tasks.send_to_a_customer'],
                         target='sent'),
        Transition(action_name='void', sources=['draft', 'paid'], target='voided'),
    ]


from django.db import transaction
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST

from django_logic.exceptions import TransitionNotAllowed


class ViewSetActionMixin(object):
    def perform_action(self):
        action_name = kwargs.get('action_name')

        if self.proxy_model is None:
            raise FSMException('Please, provide proxy_model')

        if not action_name:
            return Response(status=HTTP_400_BAD_REQUEST, data={
                "errors": ["Action not specified".format()]
            })

        instance = self.proxy_model.objects.get(pk=self.get_object().pk)

        if action_name not in get_available_user_transitions(instance, self.action_field, request.user):
            return Response(status=HTTP_400_BAD_REQUEST, data={
                "errors": ["Action '{}' is not allowed".format(action_name.capitalize())]
            })

        with transaction.atomic():
            instance = self.proxy_model.objects.select_for_update().get(pk=instance.pk)
            try:  # with provided parameters
                method_to_call = getattr(instance, action_name)
                method_to_call(**request.data)
            except TypeError as er:
                raise ValidationError(er)


            instance.save()
            serializer = self.get_serializer(instance)
            return Response(serializer.data)

    def save_and_perform_action(self, request, action_name, pk=None):
        """
        This method allows to save the serialized data and perform a transition within one transaction

        :param request: HttpRequest
        :param action_name: str
        :param pk: int or None
        :return: Model Instance
        """
        # TODO: add hints
        try:
            with transaction.atomic():
                if pk is None:
                    serializer = self.get_serializer(data=request.data)
                    serializer.is_valid(raise_exception=True)
                    self.perform_create(serializer)
                    instance = serializer.instance
                else:
                    instance = self.get_object()
                    serializer = self.get_serializer(instance, data=request.data)
                    serializer.is_valid(raise_exception=True)
                    self.perform_update(serializer)

                instance = self.proxy_model.objects.select_for_update().get(pk=instance.pk)

                method_to_call = getattr(instance, action_name)
                method_to_call()
                instance.save()
        except TransitionNotAllowed:
            return Response(status=HTTP_400_BAD_REQUEST, data={
                "errors": ["Action '{}' is not allowed".format(action_name.capitalize())]
            })
        else:
            serializer = self.get_serializer(instance)
            return Response(serializer.data)



from django.apps import apps
from django.db import transaction

from django_logic.commands import SideEffects, Callbacks


class TransitionTaskFailed(Exception):
    pass


try:
    from celery import signature, group, chain, shared_task
    from celery.result import AsyncResult
except ImportError:
    pass  # TODO: handle
else:
    @shared_task(acks_late=True)
    def complete_transition(*args, **kwargs):
        app = apps.get_app_config(kwargs['app_label'])
        model = app.get_model(kwargs['model_name'])
        instance = model.objects.get(id=kwargs['instance_id'])
        transition = kwargs['transition']

        transition.complete_transition(instance, kwargs['field_name'])


    @shared_task(acks_late=True)
    def fail_transition(task_id, *args, **kwargs):
        try:
            transition = kwargs['transition']
            try:
                # Exception passed through args
                exc = args[0]
            except IndexError:
                task = AsyncResult(task_id)
                exc = task.info

            app = apps.get_app_config(kwargs['app_label'])
            model = app.get_model(kwargs['model_name'])
            instance = model.objects.get(id=kwargs['instance_id'])
            transition.fail_transition(instance, kwargs['field_name'])
        except Exception:
            # TODO: add logger
            print('Exception')


class SideEffectTasks(SideEffects):
    def execute(self, instance: any, field_name):
        if not self.commands:
            return super(SideEffectTasks, self).execute(instance, field_name)

        task_kwargs = dict(app_label=instance._meta.app_label,
                           model_name=instance._meta.model_name,
                           instance_id=instance.pk,
                           field_name=field_name)

        header = [signature(task_name, kwargs=task_kwargs) for task_name in self.commands]
        header = chain(*header)
        task_kwargs.update(dict(transition=self.transition))
        body = complete_transition.s(**task_kwargs)
        tasks = chain(header | body).on_error(fail_transition.s(**task_kwargs))
        transaction.on_commit(tasks.delay)


class CallbacksTasks(Callbacks):
    def execute(self, instance, field_name):
        if not self.commands:
            return super(CallbacksTasks, self).execute(instance, field_name)

        task_kwargs = dict(app_label=instance._meta.app_label,
                           model_name=instance._meta.model_name,
                           instance_id=instance.pk,
                           field_name=field_name)

        tasks = [signature(task_name, kwargs=task_kwargs) for task_name in self.commands]
        transaction.on_commit(group(tasks))



try:
    from celery import Celery
except ImportError:
    pass # TODO: handle it
else:
    app = Celery('tests')
    app.config_from_object('django.conf:settings', namespace='CELERY')
    app.autodiscover_tasks()



import re


def convert_to_snake_case(name):
    # TODO: test
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def convert_to_readable_name(name):
    # TODO: test
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1 \2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1)
